import os

from pathlib import Path
from typing import Union, List, Dict, Any, Optional, Tuple, Callable, Iterable

from .internals import read_any, format_dict

def read(
    path: Union[str, Path],
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """Read a file or directory of files.
    
    Supported file formats: .json, .yaml, .toml, .proto
    """
    path = Path(path)
    
    if path.is_dir():
        return read_dir(path)
    else:
        return read_file(path)
    
def read_file(
    path: Union[str, Path],
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """Read a file."""
    path = Path(path)
    
    suffix = path.suffix
    reader = format_dict["readers_compact"].get(suffix, None)
    
    if reader is not None:
        return reader(path)
    else:
        return read_any(format_dict["readers"], path, msg=f'File format {suffix} not supported.')
        
def read_proto(
    path: Union[str, Path],
    ) -> Dict[str, Any]:
    """Read a proto file."""
    raise NotImplementedError('Proto file reading not implemented yet.')
  
def read_dir(
    path: Union[str, Path],
    ) -> List[Dict[str, Any]]:
    """Read a directory of files."""
    readfiles = []
    for file in os.listdir(path):
        if any(file.endswith(ext) for ext in format_dict["formats"]):
            readfiles.append(read_file(file))
            
    return readfiles


import os
import sys
import json
import rich
import typer
import subprocess
from signal import SIGINT, signal
from rich.prompt import Prompt, Confirm


from iotree.core.io.reader import read
from iotree.core.io.writer import write

from iotree.core.render.trees import (
    build
)

from iotree.core.render.demo import (
    print_demo, demo_symbols, demo_themes,
    colorTable, themeTable, symbolTable, extrasTable,
    render_file_demo
)

from iotree.core.render.tables import (
    tableFromRecords, treeThemeToTable, recordFormatting,

)
from iotree.utils.paths import (
    config_dir, package_dir,
    base_dir, tests_dir, safe_config_load
)

console = rich.console.Console()

symbols, themes, user_info, local_config = safe_config_load()

app = typer.Typer(
    name='iotree',
    help='A CLI & python package for rendering markup language docs as trees.',
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode='rich',
    )

config = typer.Typer(
    name='config',
    help='Configure the style of your tree.',
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode='rich',
    )

dev = typer.Typer(
    name='develop',
    help="Work ☕💻 on the package's 📦 content.",
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode='rich',
    )

view = typer.Typer(
    name='view',
    help="View 🔎 the package's 📦 themes 🖌️ & symbols ➡️ for customization 💅",
    no_args_is_help=True,
    rich_help_panel="rich",
    rich_markup_mode='rich',
    )

app.add_typer(config, name='config',
    help='Configure the ✨ style ✨ of your tree. [yellow]Alias: `cfg`.[/yellow]',
    rich_help_panel="rich",
    )

app.add_typer(config, name='cfg', hidden=True,
    help='Alias for `config`. Configure the ✨ style ✨ of your tree.',
    rich_help_panel="rich",
    )

app.add_typer(view, name='view', hidden=False,
    help="View 🔎 the package's 📦 themes 🖌️ & symbols 🗼 for customization 💅",
    rich_help_panel="rich",
    )

app.add_typer(dev, name='develop', hidden=False,
    help="Work 💻 on the package's 📦 content.",
    rich_help_panel="rich",
    )

@app.command(name='render', help='Render a Markup Language file 📂 as a tree.')
def render(
    file = typer.Argument(..., help='The file to render as a tree.'),
    ):
    """Render a Markup Language file 📂 as a tree 🌳."""
    obj = read(file)
    console.print(build(obj))

@dev.command(name='check', help='Run checks ❓ on the package 📦')
def checks():
    """Run checks on the package."""
    os.system('pytest')
    
@app.command(name='demo', help='Run a demo of the CLI part of the package.')
def demo():
    """Run a demo of the CLI part of the package."""
    render_file_demo(appname='iotree')

@config.command(name='init', help='Initialize a config file.')
def initialize(
    user: str = typer.Option(None, "-u", "--user", help='The user to set the value for.'),
    use_saved: bool = typer.Option(True, "-us", "--use-saved", help='Use the saved config.'),
    ):
    """Set a config value. These values are saved in the config directory.
    
    The file is named `local-config.json` can be updated either manually or through this command.
    """
    user = (
        user_info['last_user'] if user is None
        else user or os.getlogin()
        )
    
    ok = None
    console.print(f'[bold orange]Warning: no user declared.[/][dim magenta] using default/declared value = [underline]{user}[/][/]')
    if use_saved and len(local_config) > 0:
        lconf = local_config["user_info"] if "user_info" in local_config else local_config
        if user not in lconf:
            lconf["user_info"] = {user: "default"}
            console.print(f'[bold magenta]No config found for user [underline]{user}, [/][/][bold magenta]using default value.[/]')
        else:
            uconf = lconf["user_info"][user]
            uconf_theme = user_info[user]['theme'] if user in user_info else user_info['default']['theme']
            console.print(f'[bold magenta]Config found for user {user}[/]')
            console.print(f'[dim magenta]Config: {uconf} ===> {uconf_theme}: {themes[uconf_theme]}[/]')
            ok = Confirm.ask('[bold green underline]No need to initialize.[/] [bold magenta] Would you still like to overwrite these ?[/]')
    if ok is None:
        ok = Confirm.ask('[bold red]No saved config found.[/] [bold magenta] Would you like to create one?[/]')
    if ok:
        lconf = {}
        choices = demo_symbols()
        user_symbol = Prompt.ask(
            '[bold magenta]Which [bold yellow]symbol[/bold yellow] did you [bold pink]like most[/bold pink] ?[/bold magenta]',
            choices=choices, show_choices=True
            )
        user_data = {"name":user, "symbol": user_symbol}
        choices = demo_themes()
        user_theme = Prompt.ask(
            '[bold magenta]Which [bold yellow]theme[/bold yellow] did you [bold pink]like most[/bold pink] ?[/bold magenta]',
            choices=choices, show_choices=True
            )
        user_data['theme'] = user_theme
        user_info[user] = user_data
        lconf = { "user_info": {user: user_theme}, "last_user": user }
        write(user_info, config_dir /'user-info.json')
        write(lconf, config_dir /'user-info.json')
    else:
        console.print('[dim yellow]Proceeding with default values.[/]')
        lconf = { "user_info": {user: "default"}, "last_user": user }
        write(lconf, config_dir /'user-info.json')
            
@config.command(name='set', help='Set a config value.', no_args_is_help=True)
def setter(
    param: str = typer.Argument(..., help='The parameter to set.'),
    value: str = typer.Argument(..., help='The value to set the parameter to.'),
    user: str = typer.Argument(None, help='The user to set the value for.'),
    ):
    """Set a config value. These values are saved in the config directory.
    
    Possible values for `param` are:
    - `symbol`: The symbol to use for the tree.
    - `theme`: The theme to use for the tree.
    - `last_user`: The default user whose profile will be loaded by default.
    
    [bold red]Note: [/]The `last_user` parameter is set automatically when the `init` command is run.
    [bold red]Note: [/][orange]The value is [underline]only set for the given user[/underline][/].
    """
    confpaths = [config_dir / 'local-config.json', config_dir / 'user-settings.json']

    __local_config = read(confpaths[0])
    __user_info = read(confpaths[1])

    param, value = param.lower(), value.lower()

    if param == 'last_user':
        __local_config['last_user'] = value
        write(__local_config, confpaths[0])
        console.print(f'[bold magenta] ✅ Config found for user {user}[/]')
        console.print(f'[dim magenta] 👷 Config: {param} ===> {value}[/]')
        sys.exit(0)

    user = user.lower() if user is not None else None

    if user is not None and user not in __user_info:
        ok = Confirm.ask(f'[bold red] ❌ No config found for user {user}[/] [bold magenta] Would you like to create one?[/]')
        if ok:
            __user_info[user] = {}
            if param == 'theme':
                if value in themes:
                    __user_info[user]['theme'] = value
                    __local_config['user_info'] = {user: value}
                else:
                    __user_info[user]['theme'] = 'default'
                    console.print(f'[bold red] ❌ No theme found for user {user}[/] [bold magenta] Using default theme.[/]')
            elif param == 'symbol':
                if value in symbols:
                   __user_info[user]['symbol'] = value
                else:
                    __user_info[user]['symbol'] = '{'
                    console.print(f'[bold red] ❌ No symbol found for user {user}[/] [bold magenta] Using default symbol.[/]')
            write(__user_info, confpaths[1])
            console.print(f'[bold magenta] ✅ Config created for user {user}[/]')
            console.print(f'[dim magenta] 👷 Config: {param} ===> {value}[/]')
            write(__user_info, confpaths[1])
            write(__local_config, confpaths[0])
            sys.exit(0)
        else:
            console.print(f'[bold red] ❌ No config found for user {user}[/]')
            sys.exit(0)

    user = (
        user if user is not None
        else local_config['last_user'] if 'last_user' in local_config
        else os.getlogin()
        )
    
    for i, conf in enumerate([__local_config, __user_info]):
        uconf = (
            conf[user] if user in conf
            else conf["user_info"][user] if "user_info" in conf and user in conf["user_info"]
            else conf["user_info"] if "user_info" in conf
            else conf
            )
        if param in uconf:
            uconf[param] = value
            json.dump(
                conf, open(confpaths[i], 'w+'), indent=4
                )
            console.print(f'[dim magenta]Config: {uconf} ===> {param}: {value}[/]')
            sys.exit(0)
    console.print(f'[bold red] ❌ No config found for user {user}[/]')

@config.command(
    name='from-file',
    help='Add a new user from a JSON 📁 user entry. [bold yellow] Alias: `ff`[/]',
    no_args_is_help=True
    )
def from_file(
    markup_file: str = typer.Argument(..., help='The JSON/TOML/XML/YAML file containing the user entry.'),
    ):
    """Add a new user from a JSON 📁 user entry.
    
    The JSON file should have the following structure:
    ```json
    {
        "user": "<the user name>",
        "name": "<your first name>",
        "symbol": "<the symbol you like most>",
        "theme": "<the color theme you prefer>"
    }
    ```
    The equivalent TOML file would be:
    
    ```toml
    user = "<the user name>"
    name = "<your first name>"
    symbol = "<the symbol you like most>"
    theme = "<the color theme you prefer>"
    ```
    
    [bold red]Note: [/]The `last_user` parameter is set automatically when the `init` command is run.
    This command will add the user to the `user_info` dictionary in the `user-settings.json` file,
    but not set the last user in the `local-config.json` file.
    """
    mandatory_keys = ['user', 'name', 'symbol', 'theme']
    content = read(markup_file)
    if len(set(mandatory_keys).difference(content.keys())) > 0:
        str_mandatory_keys = '\n- '.join(mandatory_keys)
        console.print(
            f"""[bold red]The Markup file is missing one of the following keys:\n- {str_mandatory_keys}[/]"""
        )
        sys.exit(1)
    else:
        user_info[content['user']] = content
        local_config['user_info'][content['user']] = content['theme']
        json.dump(user_info, open(config_dir / 'user-settings.json', 'w+'), indent=4)
        json.dump(local_config, open( config_dir /'local-config.json', 'w+'), indent=4)
        console.print(f'[bold green]Added user {content["user"]} to config.[/]')

@config.command(
    name='ff',
    help='Alias for `from-file`, adds user from Markup language file',
    no_args_is_help=True, hidden=True
    )
def ff(
    markup_file: str = typer.Argument(..., help='The JSON/TOML/XML/YAML file containing the user entry.'),
    ):
    from_file(markup_file)

@config.command(name='get', help='Get a config value.', no_args_is_help=True)
def getter(
    param: str = typer.Argument(..., help='The parameter to get.'),
    user: str = typer.Option(None, help='The user to get the value for.'),
    ):
    confpaths = [config_dir / 'local-config.json', config_dir / 'user-settings.json']
    
    __local_config = read(confpaths[0])
    __user_info = read(confpaths[1])

    user = (
        user if user is not None
        else os.getlogin()
        )

    param, user = param.lower(), user.lower()

    if param == 'last_user':
        value = __local_config['last_user']
        console.print(f'[bold magenta] ✅ Config found for user {user}[/]')
        console.print(value)
        sys.exit(0)
    
    if param == "user":
        if user in __local_config["user_info"] or user in __user_info:
            value = f"user: {__user_info[user]}" if user in __user_info else f"user: {__local_config['user_info'][user]}"
            console.print(f'[bold magenta] ✅ Config for themes found for user {user}[/]')
            console.print(value)
            sys.exit(0)
        else:
            console.print(f'[bold red] ❌ No config found for user {os.getlogin()}[/]')
            console.print(f'[dim yellow] ❗️ You can add a config for {os.getlogin()} by running `config from-file <file>` or `config init`, or `config set <param> <value> <user>`[/]')
            sys.exit(1)

    user = os.getlogin() if "last_user" not in __local_config else __local_config["last_user"]
    
    for i, conf in enumerate([__local_config, __user_info]):
        uconf = (
            conf[user] if user in conf
            else conf["user_info"][user] if "user_info" in conf and user in conf["user_info"]
            else conf["user_info"] if "user_info" in conf
            else conf
            )
        if param in uconf:
            value = uconf[param]
            console.print(f'[bold magenta] ✅ Config found for user {user}[/]')
            console.print(value)
            sys.exit(0)
    console.print(f'[bold red] ❌ No config found for user {user}[/]')

@config.command(name='list', help='List all config values. [bold yellow] Alias: `ls`[/]')
def lister():
    rows = []
    for user, values in user_info.items():
        values['user'] = user
        rows.append(values)
    
    table = tableFromRecords(rows, title='User Settings', theme=user_info['default']['theme'])
    console.print(table)
    
@config.command(name='ls', help='Alias for `list`.', hidden=True)
def ls():
    lister()

@config.command(name='reset', help='Reset the config file.')
def reset():
    """Reset the config file."""
    os.remove(local_config)
    
    
@view.command(
    name='colors',
    help='View currently available 🏳️‍🌈 color 🏳️‍🌈 options',
    )
def view_colors():
    """View currently available 🏳️‍🌈 color 🏳️‍🌈 options"""
    console.print(
            colorTable()
        )

@view.command(
    name='themes',
    help='View currently available 🖌️ theme 🖌️ options',
    )
def view_themes():
    """View currently available 🖌️ theme 🖌️ options"""
    console.print(
            themeTable()
        )

@view.command(
    name='symbols',
    help='View currently available 🗼 symbols 🗼 options',
    )
def view_themes():
    """View currently available 🗼 symbols 🗼 options"""
    console.print(
            symbolTable()
        )

@view.command(
    name='extras',
    help='View extra 🎁 symbols 🎁 options 🧐 [dim](not included in the default set)[/]',
    )
def extras(num: int = typer.Option(100, "-n", "--num", help='The number of extra symbols to view.')):
    """View extra 🎁 symbols 🎁 options 🧐 [dim](not included in the default set)[/]"""

    rich.print("""
[bold magenta]I had to keep the basic symbols quite light, however,[/]
    
[bold green]There are [underline]a lot of other symbols[/] available.[/]
If you want to add one to the list, feel free to either [bold cyan]contact me[/]
or add the corresponding unicode to the symbols file [yellow]using the `extra` config command.
[/]
    """)
    ok = typer.confirm(f'You are about to view {num} extra symbols. Do you want to continue?', default=True)
    if ok:
        console.print(
            extrasTable(num)
        )
    else:
        console.print('[bold red]Aborting.[/]')
        console.print('[dim]You can always view the extra symbols later.[/]')
        console.print('[green]Tip: Use the -n or --num parameter[/]')
        console.print('[bold yellow]Exiting...[/]')
        sys.exit(0)
   


import rich
from typing import Any, Dict, List, Optional, Union


from iotree.utils.paths import (
    package_dir, base_dir, config_dir, safe_config_load
)

symbols, themes, user_infos, local_config = safe_config_load()

def treeThemeToTable(theme: Union[str, Dict[str, Any]] = "default"):
    """Convert a theme fit for a tree to one for a rich.table.Table object."""
    if isinstance(theme, str):
        theme = themes[theme]
    table_theme = {
        "header": theme["node"][0],
        "rows": theme["node"][1:3],
    }
    return table_theme

def recordFormatting(record: Dict[str, Any], columns: List[str]) -> Dict[str, Any]:
    """Add empty column values to a partial record + convert values to strings.
    
    Also orders the columns in the same order as the columns list."""
    lrow = []
    for col in columns:
        if col not in record.keys():
            lrow.append("")
        else:
            lrow.append(str(record[col]))
    return lrow

def tableFromRecords(
    records: List[Dict[str, Any]],
    title: Optional[str] = "Table of records",
    theme: Optional[str] = None,
    ) -> rich.table.Table:
    theme = treeThemeToTable(theme)
    tab = rich.table.Table(
        title=title,
        header_style=theme["header"],
        box=rich.box.ROUNDED,
        row_styles=theme["rows"]
        )
    columns = list(records[0].keys())
    for col in columns:
        tab.add_column(col, justify="center")
    for record in records:
        tab.add_row(*recordFormatting(record, columns))
    return tab


import orjson
import yaml
import toml
import soup2dict as std
import xmltodict as xtd

from bs4 import BeautifulSoup
from typing import Union, List, Dict, Any, Optional, Tuple, Callable, Iterable

def __call__any(funcs: List[Callable], *args, **kwargs) -> Any:
    """Try all callables in a list until one succeeds."""
    for f in funcs:
        try:
            return f(*args, **kwargs)
        except Exception as e:
            continue
    if "msg" in kwargs:
        raise ValueError(f'All functions failed. {kwargs["msg"]}')
    else:
        msg = [
            f"\n - {funcs[0].__name__} failed with args: {args} and kwargs: {kwargs}" for f in funcs
        ]
        msg = "f'All calls failed:" + "".join(msg)
        raise ValueError(msg)

orjson_options = orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SERIALIZE_DATACLASS | orjson.OPT_SERIALIZE_UUID | orjson.OPT_NAIVE_UTC | orjson.OPT_INDENT_2 

format_dict = {
    "formats": ['.json', '.yaml', '.toml', '.xml', '.html'],
    "readers": [
        lambda path : orjson.loads(open(path,'rb').read()),
        lambda path : yaml.safe_load(open(path, 'r')),
        lambda path : toml.loads(open(path, 'r').read()),
        lambda path : xtd.parse(open(path, 'r').read()),
        lambda path : std.convert(BeautifulSoup(open(path, 'r').read(), 'html.parser'))
    ],
    "writers": [
        lambda data, path : open(path, 'wb+').write(orjson.dumps(data, option=orjson_options)),
        lambda data, path : yaml.safe_dump(data, open(path, 'w+')),
        lambda data, path : toml.dump(data, open(path, 'w+')),
        lambda data, path : open(path, "w+").write(xtd.unparse(data)),
        lambda data, path : open(path, "w+").write(std.convert(data, as_json=True))
    ],
}

format_dict["readers_compact"] = { k : v for k, v in zip(format_dict["formats"], format_dict["readers"]) }
format_dict["writers_compact"] = { k : v for k, v in zip(format_dict["formats"], format_dict["writers"]) }



def write_any(data, path) -> Any:
    """Try all writers in a list until one succeeds."""
    return __call__any(format_dict["writers"], data, path)

def read_any(funcs: List[Callable], *args, **kwargs) -> Any:
    """Try all readers in a list until one succeeds."""
    return __call__any(format_dict["readers"], path)


import rich

from typing import Any, Callable, List, Dict, Iterable
from rich.console import Console
from rich.progress import (
    Progress, BarColumn, TextColumn,
    TimeRemainingColumn, TimeElapsedColumn,
    SpinnerColumn
)

from iotree.utils.paths import (
    package_dir, base_dir, config_dir, safe_config_load
)

symbols, themes, user_infos, local_config = safe_config_load()


def call_any(funcs: List[Callable], *args, **kwargs) -> Any:
    """Try all functions in a list until one succeeds."""
    for f in funcs:
        try:
            return f(*args, **kwargs)
        except Exception as e:
            continue
    if "msg" in kwargs:
        raise ValueError(f'All functions failed. {kwargs["msg"]}')
    else:
        msg = [
            f"\n - {funcs[0].__name__} failed with args: {args} and kwargs: {kwargs}" for f in funcs
        ]
        msg = "f'All functions failed:" + "".join(msg)
        raise ValueError(msg)
    

def format_user_theme(theme: Dict[str,str]) -> Dict[str,str]:
    """Format a user theme to be used by rich."""
    must_have = [
        "description", "complete",
        "finished", "remaining",
        "message"]
    
    default = {
        "description": "bold purple",
        "complete": "bold green3",
        "finished": "bold green3",
        "remaining": "bold pink3",
        "message": "magenta",
        "spinner": "dots",
        "bar_width": 70,
    }
    
    for mh in must_have:
        if mh in theme:
            default[mh] = theme[mh]
        for k, v in theme.items():
            if k.endswith(mh):
                default[k] = v
    return default

def apply_progress_theme(
    theme: Dict[str, str],
    console: Console = None,
    ) -> Progress:
    """Apply a theme to the progress bar."""
    theme = format_user_theme(theme)
    
    return Progress(
                SpinnerColumn(theme["spinner"]),
                TextColumn("{task.description}", justify="right", style=theme.get("task.description", theme["description"])),
                BarColumn(bar_width=None),
                "[progress.percentage]{task.percentage:>3.1f}%",
                TimeElapsedColumn(),
                TimeRemainingColumn(),
                console=console,
            )
    

def basic_pbar(
    console: Console = None,
    ) -> Progress:
    return apply_progress_theme(theme=format_user_theme({}), console=console)

def rich_func(func, *args, **kwargs) -> Any:
    """Run a function with rich progress bar."""
    args = [], kwargs = {}
    theme = kwargs.pop("theme") if "theme" in kwargs else format_user_theme({})
    console = kwargs.pop("console") if "console" in kwargs else Console()
    pbar = kwargs.pop("progress") if "progress" in kwargs else apply_progress_theme(theme=theme, console=console)
    with pbar:
        task_id = pbar.add_task(f"Running {func.__name__}", total=1)
        try:
            result = func(*args, **kwargs)
            pbar.update(task_id, advance=1)
        except Exception as e:
            pbar.console.print(f"[bold red]Error while running {func.__name__}[/bold red]")
            pbar.console.print(e)
            pbar.update(task_id, advance=1)
            raise e
    return result
    
        
    
def rich_func_chainer(
    funcs: List[Callable], params: List[Any], *args, **kwargs
    ) -> Iterable[Any]:
    """Run a list of functions with rich progress bar.
    
    If you want to customize the progress bar, you can pass a `progress` keyword argument.
    If you want to customize the console, you can pass a `console` keyword argument.
    If you want a specific color theme or style for the progress bar, you can pass a `theme` keyword argument.
    
    Args:
        funcs (List[Callable]): A list of functions to run.
        params (List[Any]): A list of parameters to pass to each function. They must be in the same order as the functions.
        *args: Arguments to pass to each function. These will be passed to all functions.
        **kwargs: Keyword arguments to pass to each function. These will be passed to all functions.

    Returns:
        Iterable[Any]: An iterable of the results of each function.
    """

    if len(funcs) != len(params):
        if not isinstance(funcs, list):
            funcs = [funcs]*len(params)
        else:
            raise ValueError("The number of functions and parameters must be the same.")
    
    progress = kwargs.pop("progress", None)
    console = kwargs.pop("console", None)
    theme = kwargs.pop("theme", None)
    
    if console is None:
        console = Console()
    if progress is None:
        if theme is None:
            progress = Progress(
                SpinnerColumn(),
                TextColumn("[bold purple]{task.description}", justify="right"),
                BarColumn(bar_width=None),
                "[progress.percentage]{task.percentage:>3.1f}%",
                TimeElapsedColumn(),
                TimeRemainingColumn(),
                console=console,
            )

        else:
            progress = apply_progress_theme(theme=theme, console=console)

    errs = []

    with progress:
        for i, f in enumerate(funcs):
            task_id = progress.add_task(f"Running {f.__name__}", total=1)
            try:
                if isinstance(params[i], list):
                    result = f(*params[i], *args, **kwargs)
                elif isinstance(params[i], dict):
                    result = f(*args, **params[i], **kwargs)
                else:
                    result = f(params[i], *args, **kwargs)
                    
                progress.update(task_id, advance=1)
            except Exception as e:
                progress.console.print(f"[bold red]Error while running {f.__name__}[/bold red]")
                progress.console.print(e)
                progress.update(task_id, advance=1)
                errs.append(e)
                result = None
            yield result
    if errs:
        fmt_errs = '\n - '.join([str(e) for e in errs])
        raise ValueError(f"Errors while running functions: {fmt_errs}")


from .funcs import (
    call_any, rich_func, rich_func_chainer,
    format_user_theme, apply_progress_theme,
)

from .tables import (
    tableFromRecords, treeThemeToTable, recordFormatting,
    
)

from .trees import (
    build
)

from .theme import (
    initConfig, convertTheme, check_none, lnode
)

from .demo import (
    print_demo, demo_symbols, demo_themes,
    colorTable, themeTable, symbolTable, extrasTable,
    render_file_demo
)

from ..io import reader, writer, serialize




import rich
from time import sleep

from rich.tree import Tree
from itertools import cycle
from typing import (
    Any, Dict, List, Union,
    Optional, Tuple, MutableMapping,
    MutableSequence
    )


from iotree.core.render.theme import (
    initConfig, convertTheme, check_none, lnode
)

from iotree.core.render.funcs import (
    format_user_theme, apply_progress_theme,
)

from iotree.utils.paths import (
    package_dir, base_dir, config_dir, safe_config_load
)

symbols, themes, user_infos, local_config = safe_config_load()


def build(
    dictlike: Union[Dict[str, Any], List[Dict[str, Any]]],
    state: Optional[Tree] = None,
    theme: Optional[Union[str,Dict[str,str]]] = None,
    symbol: Optional[str] = None,
    user: Optional[str] = None,
    numbered: Optional[bool] = False,
    depth: Optional[int] = 10,
    cyclcols: Optional[cycle] = None,
    ) -> Tree:
    """Build a tree from a dictionary or list of dictionaries.
    
    The dictionary is the result of reading a file or directory of files.
    Supported file formats: .json, .yaml, .toml
    
    Args:
        dictlike (Union[Dict[str, Any], List[Dict[str, Any]]]): A dictionary or list of dictionaries.
        state (Optional[Tree], optional): A rich.tree.Tree object. Defaults to None. Is used for recursion.
        theme (Optional[Dict[str,str]], optional): A theme from the themes.json file. Defaults to 'default'.
        symbol (Optional[str], optional): A symbol from the symbols.json file. Defaults to 'star'.
        user (Optional[str], optional): A user from the user-settings.json file. Defaults to None.
        numbered (Optional[bool], optional): Whether or not to number the list nodes. Defaults to False.
        depth (Optional[int], optional): The depth of the tree. Defaults to 10. [bold red]Not yet implemented[/].
        cyclcols (Optional[cycle], optional): A cycle of colors. Defaults to None. Is used for recursion. 
    """
    theme = convertTheme(theme)
    
    if state is None:
        __theme, __symbol = initConfig(user)
        theme = __theme if check_none(theme) else theme
        symbol = __symbol if symbol is None else symbol
        node_colors = theme['node']
        leaf_color = theme['leaf']
        cyclcols = cycle(node_colors)
        _col = next(cyclcols)
        state = Tree(f'[bold {_col}]Contents[/]')
    else:
        __theme, __symbol = initConfig(user)
        theme = __theme if check_none(theme) else theme
        symbol = __symbol if symbol is None else symbol
        node_colors = __theme['node']
        leaf_color = __theme['leaf']
        
    if symbol in ['par', 'curl', 'sqbra', 'dbcol', 'less']:
        numbered = True
    else:
        numbered = False
        
    if isinstance(dictlike, (list, MutableSequence)):
        _col = next(cyclcols)
        
        if numbered:
            i = 0
        for d in dictlike:
            if not isinstance(d, (dict, list)):
                state.add(f"[{leaf_color}]{d}[/]")
            else:
                if numbered:
                    _symbol = lnode(symbol, index=i)
                    i += 1
                else:
                    _symbol = lnode(symbol)
                branch = state.add(f"[bold {_col}]{_symbol}[/]")
                build(d, state=branch,
                      cyclcols=cyclcols,
                      depth=depth-1,
                      theme=theme,
                      numbered=numbered,
                      symbol=symbol,
                      user=user)
    elif isinstance(dictlike, (dict, MutableMapping)):
        _col = next(cyclcols)
        for key, value in dictlike.items():
            if not isinstance(value, (dict, list)):
                state.add(f"[bold {_col}]{key}[/]:[{leaf_color}] {value}[/]")
            else:
                branch = state.add(f"[bold {_col}]{key}:[/]")
                build(value, state=branch,
                      cyclcols=cyclcols,
                      depth=depth-1,
                      theme=theme,
                      numbered=numbered,
                      symbol=symbol,
                      user=user
                      )
    else:
        if not isinstance(dictlike, (str, float, int)):
            raise TypeError(f"Expected a dictionary or list of dictionaries, got {type(dictlike)}")
        state.add(f"[{leaf_color}] {dictlike}[/]")
    return state


import json
from pathlib import Path
from typing import Dict, Any, Tuple, Union, List, Callable

utils_dir = Path(__file__).parent
package_dir = utils_dir.parent
base_dir = package_dir.parent
tests_dir = base_dir / 'tests'
config_dir = package_dir / 'config'

def safe_json_load(path: Union[str, Path]) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """Load a file or directory of files."""
    path = Path(path)
    if path.exists():
        if path.is_dir():
            return [json.load(open(file, 'r')) for file in path.iterdir()].pop()
        else:
            return json.load(open(path, 'r'))
    else:
        return {}

def safe_config_load() -> Tuple[Dict[str, Any]]:
    """Load config files.
    
    In this specific order:
    - symbols.json
    - themes.json
    - user-settings.json
    - local-config.json
    """
    sym_path = config_dir / 'symbols.json'
    themes_path = config_dir / 'themes.json'
    uinfo_path = config_dir / 'user-settings.json'
    lconf_path = config_dir / 'local-config.json'
    
    symbols = safe_json_load(sym_path)
    themes = safe_json_load(themes_path)
    user_infos = safe_json_load(uinfo_path)
    local_config = safe_json_load(lconf_path)
    
    return symbols, themes, user_infos, local_config



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['iotree',
 'iotree.cli',
 'iotree.core',
 'iotree.core.io',
 'iotree.core.render',
 'iotree.utils']

package_data = \
{'': ['*'], 'iotree': ['config/*', 'examples/*']}

install_requires = \
['httpx>=0.22.0,<0.23.0',
 'orjson>=3.8.5,<4.0.0',
 'pytest>=7.2.0,<8.0.0',
 'pyyaml>=6.0,<7.0',
 'rich>=10.11.0,<13.0.0',
 'soup2dict>=2.1.0,<3.0.0',
 'toml>=0.10.2,<0.11.0',
 'typer[all]>=0.7.0,<0.8.0',
 'xmltodict>=0.13.0,<0.14.0']

entry_points = \
{'console_scripts': ['iotree = iotree.__main__:app']}

setup_kwargs = {
    'name': 'iotree',
    'version': '0.1.11',
    'description': 'A lightweight CLI + lib that allows you to perform basic IO tasks and display your content as rich trees and tables.',
    'long_description': '# A many-in-one tool for managing your Markup Language files.\n\n## What is it?\n\n**iotree** is a tool for managing your Markup Language files. It is capable to write and read files in the following formats:\n\n- JSON\n- YAML\n- TOML\n- XML\n- And soon more... 😉\n\nThe basic goal was to have a small package anyone could add to their project and use it to manage their files. It is also possible to use it as a CLI tool.\n\n## Installation\n\nYou cannot install the CLI tool separately for now. You can install it with the following command:\n\n```bash\npip install iotree\n```\n\n## Usage\n\n### As a CLI tool\n\nTo see what the display function can do, you can use the following command:\n\n```bash\niotree demo\n```\n\nFor example, the following JSON file (displayed in VSCode)\n\n```json\n{\n    "glossary": {\n        "title": "example glossary",\n        "GlossDiv": {\n            "title": "S",\n            "GlossList": {\n                "GlossEntry": {\n                    "ID": "SGML",\n                    "SortAs": "SGML",\n                    "GlossTerm": "Standard Generalized Markup Language",\n                    "Acronym": "SGML",\n                    "Abbrev": "ISO 8879:1986",\n                    "GlossDef": {\n                        "para": "A meta-markup language, used to create markup languages such as DocBook.",\n                        "GlossSeeAlso": [\n                            "GML",\n                            "XML"\n                        ]\n                    },\n                    "GlossSee": "markup"\n                }\n            }\n        }\n    }\n}\n```\n\nwill be displayed like this:\n\n![JSON file displayed](https://i.imgur.com/tUSyW3L.png)\n\nWhile the following YAML file (displayed in VSCode)\n\n```yaml\t\n---\n doe: "a deer, a female deer"\n ray: "a drop of golden sun"\n pi: 3.14159\n xmas: true\n french-hens: 3\n calling-birds:\n   - huey\n   - dewey\n   - louie\n   - fred\n xmas-fifth-day:\n   calling-birds: four\n   french-hens: 3\n   golden-rings: 5\n   partridges:\n     count: 1\n     location: "a pear tree"\n   turtle-doves: two\n```\n\nwill be displayed like this:\n\n![YAML file displayed](https://i.imgur.com/t3q5yHS.png)\n\n**Note**: The CLI tool is not yet finished. It is still in development.  \nIf this just looks like a wrapper around [rich trees](https://rich.readthedocs.io/en/stable/tree.html)) to you, it almost because it is. :wink:\n\nAs a CLI tool, the key difference I want to bring is the ability to configure *themes* and *styles*.\n\nJust run the following command to interactively create a theme:\n\n```bash\niotree config init\n```\n\nBut if you\'re lazy, just use a file:\n\n```bash\niotree config init from-json my_theme.json\n```\n\nFor example, the following JSON file\n\n```json\n{   \n    "name": "My super pseudonym",\n    "username": "my.username",\n    "symbol": "lgpoint",\n    "theme": "bright-blue-green"\n}\n```\n\nDoes _not_ immediately affect the display of the files. You need to use the following command:  \nTo understand why, you can run\n\n```bash\niotree config ls # same as iotree cfg ls\n```\n\nThis is what the output looks like:\n\n![iotree config ls](https://i.imgur.com/5R3IMUG.png)\n\nYou can see that the theme is registered, but will **not** be used by default.\nThe best proof of this is to ask who the current user is:\n\n```bash\niotree cfg get last_user\n#> ✅ Config found for user arnov\n#> arnov\n```\nSo the current user is `arnov`. To use _YOUR_ theme, you need to run the following command:\n\n```bash\niotree cfg set last_user my.username\n#> 👷 Config: last_user ===> my.username\n```\n\nNow, if you try to render a file, you will see that the theme is used:\n\n```bash\niotree render .\\iotree\\examples\\example.toml\n```\n\nHere\'s what the output looks like:\n\n![iotree render after](https://i.imgur.com/RHROUWg.png)\n\nThe theme is now used. 🎉🎉\n\n### As a Python package\n\nYou can use the package in your Python project. Here\'s an example:\n\n```python\nfrom iotree.core.render import build\n\nmydict = {\n    "super" : {\n        "dict": {\n            "just": {\n                "for": {\n                    "example": "🤷\u200d♂️"\n                }\n            }\n        }\n    }\n}\n\n# Build the tree from the dict without hassle\n\ntree = build(mydict) # or build(mydict, theme="my_style", symbol="my_symbol")\n\n# Display the rich tree later\n\nconsole.print(tree) # or rich.print(tree)\n```\n\nBut there is more. You can also use the \'all in one\' IO from the package:\n\n```python\nfrom iotree.core.io import reader, writer\n\n# Read a file, any extension is supported from the list\n\ndict_one = reader.read("my_file.json")\ndict_two = reader.read("my_file.yaml")\ndict_three = reader.read("my_file.toml")\ndict_four = reader.read("my_file.xml")\ndict_five = reader.read("my_file.html")\n\n# ... do some manipulations, like possibly merging the dicts or creating a new one\n\n# Write the dict to a file\n\nwriter.write("my_file.json", dict_one)\nwriter.write("my_file.yaml", dict_two)\nwriter.write("my_file.toml", dict_three)\nwriter.write("my_file.xml", dict_four)\n# writer.write("my_file.html", dict_five) # Not supported yet\n```\n\nAnother small feature is the ability to _detect the type of a file_:\n\n```python\ndict_guess = reader.read("mysterious_file.not_a_real_extension")\n# if the content of the file is one of the supported types, it will be read correctly\n```\n\nThere\'s also a functional package, which is not yet documented.  \nIt allows you to compose functions while displaying rich progress using `rich.progress`.\n\n## Contributing\n\nIf you want to contribute, you can do so by forking the repository and creating a pull request.\n\nYou can also help me pick the new themes and symbols.\n\nTo see the possibilities, just run either of the following commands:\n\n```bash\niotree view themes # or\niotree view symbols # or\niotree view colors\n```',
    'author': 'Arno V',
    'author_email': 'bcda0276@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import rich
import orjson
from time import sleep

from rich.text import Text
from rich.tree import Tree
from rich.rule import Rule
from rich.color import Color
from rich.panel import Panel
from rich.layout import Layout, RowSplitter, ColumnSplitter
from typing import Any, Dict, List, Union, Optional, Tuple

from ..render.trees import build
from ..io.reader import read
from ..io.serialize import dumps
from ..io.internals import format_dict
from ...utils.paths import (
    package_dir, base_dir, config_dir, safe_config_load,
)

symbols, themes, user_infos, local_config = safe_config_load()

##########################################
############ DEMO UTILS ##################
##########################################

def find_obj_text_size(obj: Any, font_size: Optional[int] = 10) -> Tuple[int, int]:
    """Find the width of the text representation of an object."""
    txt = dumps(obj)
    height = len(txt.splitlines())*10
    width = max( len(txt) for txt in txt.splitlines() )*font_size
    return height, width

def find_terminal_size() -> Tuple[int, int]:
    """Find the size of the terminal."""
    tsize = os.get_terminal_size()
    return tsize.lines, tsize.columns

def find_text_ratio(obj: Any) -> float:
    """Find the ratio of the text representation of an object to the terminal size."""
    height, width = find_obj_text_size(obj)
    theight, twidth = find_terminal_size()
    return height/theight, width/twidth

def render_file_demo(
    appname: str,
    formats: Optional[List[str]] = format_dict["formats"],
    secs: float = 0.7,
    ) -> None:
    content = []

    sizes = []
    
    for fmt in formats:
        rich.print(Rule(" Extension: " + fmt + " ", style='bold cyan'))
        raw = open(package_dir.joinpath('examples', f'example{fmt}'), 'r').read()
        rraw = Text(raw.strip(), style='bold yellow', overflow='fold')
        obj = read(package_dir.joinpath('examples', f'example{fmt}'))
        robj = Text(
            dumps(obj).strip(),
            style='bold green',
            overflow='fold')
        tree = build(obj)

        content = [rraw, robj, tree]

        # here we display our demo in a layout
        layout = Layout(name=f'Extension .{fmt}', ratio=0.7)

        layout.split_row(
                    *[
                        Layout(
                            cont,
                            name=f'Cell-no-{idx}'
                        )
                        for idx,cont in enumerate(content)
                    ]
                )
    
        rich.print(layout)
        sleep(secs)
        
    
        

def print_demo(symbol: str, exlist: List[str]) -> None:
    """Print a demo of a symbol."""
    rich.print(f"If you use the [bold yellow]{symbol}[/] symbol '{symbols[symbol]}', it will look like this:")
    rich.print(build(exlist, symbol=symbol))
    
def demo_symbols(wait_each: float = 0.7) -> List[str]:
    """Print a demo of the symbols."""
    rich.print(
        rich.rule.Rule("[bold magenta] Symbols demo [/]")
    )
    rich.print("[bold magenta]Here's a [bold yellow]list[/]:[/]")
    some_symbols = [
        'star', 'atomstar', 'lgpoint',
        'less', 'par', 'curl',
        '6star']
    exlist = [
        {"name":"item one"},
        {"name":"item two"},
        {"name":"item three"}
    ]
    rich.print(exlist)
    rich.print(
        rich.rule.Rule()
    )
    sleep(3*wait_each)
    rich.print("[bold magenta]Here's what it looks like with different symbols:[/]")
    sleep(wait_each)
    for symbol in some_symbols:
        print_demo(symbol, exlist)
        sleep(wait_each)
        
    return some_symbols
        
def demo_themes(wait_each: float = 0.5) -> List[str]:
    """Print a demo of the themes."""
    rich.print(
        rich.rule.Rule("[bold magenta] Themes demo [/]")
    )
    rich.print("[bold magenta]Here's a [bold yellow]dictionary[/]:[/]")
    some_themes = ['default', 'pink', 'bright-blue-green', 'purple-blue']
    exobj = {
        "key A": "value A",
        "key B": ["item one", "item two"],
        "key C" : {
            "key C1": "value C1",
            "key C2": "value C2",
            "key C3": {
                "last key": "last value"
            }
        }
    }
    rich.print(exobj)
    rich.print(
        rich.rule.Rule()
    )
    sleep(3*wait_each)
    rich.print("[bold magenta]Here's what it looks like with different themes:[/]")
    sleep(wait_each)
    for theme in some_themes:
        rich.print(f"[bold magenta]Theme: [bold yellow]{theme}[/][/]")
        rich.print(build(exobj, theme=theme))
        rich.print(
            rich.rule.Rule()
        )
        sleep(wait_each)        
    return some_themes


def colorTable() -> None:
    """Print a color table."""
    table = rich.table.Table(
        show_header=True,
        header_style="bold magenta",
        title="Available Colors",
        border_style="magenta",
        box=rich.box.ROUNDED,)
    
    table.add_column("Color")
    table.add_column("Code")
    table.add_column("Example")
    table.add_column("Name")
    table.add_column("Syntaxe", justify="right", no_wrap=True)
    
    for idx in range(256):
        c = Color.from_ansi(idx)
        table.add_row(
            Panel("", style=f"on {c.name}"),
            str(idx),
            f"[color({idx})] I love colors[/]",
            f"[color({idx})] {c.name}[/]",
            r"\[color({" + str(idx) + r"})] My sentence \[/]",
        )
    return table



def themeTable() -> None:
    """Print a theme table.
    Themes are made of 4 colors for higher layers, and 1
    unique color for the lowest layer."""
    table = rich.table.Table(
        show_header=True,
        header_style="bold magenta",
        title="Available Themes",
        border_style="magenta",
        box=rich.box.ROUNDED,)
    
    demo_obj = {
        "key A": "value A",
        "key B": ["item one", "item two"],
        "key C" : {
            "key C1": "value C1",
            "key C2": "Finishing the demo"
            }
    }
    
    table.add_column("Node: First Layer")
    table.add_column("Node: Second Layer")
    table.add_column("Node: Third Layer")
    table.add_column("Node: Fourth Layer")
    table.add_column("Leaf: Layer for values")
    table.add_column("Theme representation")
    table.add_column("Theme name")
    
    for name in themes:
        leaf = themes[name]["leaf"]
        nodes = themes[name]["node"]
        table.add_row(
            Panel(f"{nodes[0]}", style=f"on {nodes[0]}"),
            Panel(f"{nodes[1]}", style=f"on {nodes[1]}"),
            Panel(f"{nodes[2]}", style=f"on {nodes[2]}"),
            Panel(f"{nodes[3]}", style=f"on {nodes[3]}"),
            Panel(f"{leaf}", style=f"on {leaf}"),
            Panel(build(demo_obj, theme=name), expand=False, width=40),
            "[bold magenta]" + name + "[/]"
        )
    
    return table


def symbolTable() -> None:
    """Print a symbol table."""
    table = rich.table.Table(
        show_header=True,
        header_style="bold magenta",
        title="Available Symbols",
        border_style="magenta",
        box=rich.box.ROUNDED,)
    
    table.add_column("Symbol")
    table.add_column("Code")
    table.add_column("Example")
    table.add_column("Name")
    
    for symbol in symbols:
        unisymb = symbols[symbol]
        table.add_row(
            f"[bold cyan]{symbol}[/]",
            f"[bold yellow]{unisymb}[/]",
            f"[bold yellow]{unisymb} I [/][bold green]{unisymb}love[/][pink] {unisymb} colors[/] {unisymb}",
            f"[bold red]{symbol}[/]",
        )
    return table


def extrasTable(n=500) -> None:
    """Print the tables containing the extras symbols table."""

    table = rich.table.Table(
        show_header=True,
        header_style="bold magenta",
        title="Extra Symbols",
        border_style="magenta",
        box=rich.box.ROUNDED,)
    
    table.add_column("Symbol")
    table.add_column("Code")
    table.add_column("Example")
    table.add_column("Name")

    extra_symbols = read(config_dir / "symbols-extras.json")
    
    topk_symbols = list(extra_symbols.keys())[:n]

    for symbol in topk_symbols:
        unisymb = extra_symbols[symbol]
        table.add_row(
            f"[bold cyan]{symbol}[/]",
            f"[bold yellow]{unisymb}[/]",
            f"[bold yellow]{unisymb} I [/][bold green]{unisymb}love[/][pink] {unisymb} colors[/] {unisymb}",
            f"[bold red]{symbol}[/]",
        )
    return table


import orjson

from typing import Union, Optional

orjson_options = orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SERIALIZE_DATACLASS | orjson.OPT_SERIALIZE_UUID | orjson.OPT_NAIVE_UTC | orjson.OPT_INDENT_2 


def dumps(data, codec: Optional[str] = "utf8") -> Union[str, bytes]:
    return orjson.dumps(data, option=orjson_options).decode(codec)